#include <cyg/infra/diag.h>
#include <cyg/infra/cyg_trac.h>        // tracing macros

#include <pkgconf/hal.h>
#include <cyg/io/flash.h>
#include <cyg/crc/crc.h>
//#include <cyg/error/errno.h>

#include <string.h>
#include <stdio.h>
#define BLOCK_NUMBER (128) /// fixme das ist nur vorruebergehend

// aus fconfig.c

#define CONFIG_KEY1    0x0BADFACE
#define CONFIG_KEY2    0xDEADDEAD

// 
// Layout of config data
// Each data item is variable length, with the name, type and dependencies
// encoded into the object.
//  offset   contents
//       0   data type
//       1   length of name (N)
//       2   enable sense
//       3   length of enable key (M)
//       4   key name
//     N+4   enable key
//   M+N+4   data value
//

#define CONFIG_OBJECT_TYPE(dp)          (dp)[0]
#define CONFIG_OBJECT_KEYLEN(dp)        (dp)[1]
#define CONFIG_OBJECT_ENABLE_SENSE(dp)  (dp)[2]
#define CONFIG_OBJECT_ENABLE_KEYLEN(dp) (dp)[3]
#define CONFIG_OBJECT_KEY(dp)           ((dp)+4)
#define CONFIG_OBJECT_ENABLE_KEY(dp)    ((dp)+4+CONFIG_OBJECT_KEYLEN(dp))
#define CONFIG_OBJECT_VALUE(dp)         ((dp)+4+CONFIG_OBJECT_KEYLEN(dp)+CONFIG_OBJECT_ENABLE_KEYLEN(dp))


//#include "should_come_from_configtool.h"

#include "fisfs.h"

//#include "fis.h"
#include "flash_config.h"



bool debugFisFS=false;

namespace FisFS
{

int startUpdateDirectory();

int finishUpdateDirectory(bool success);

struct FileData
{
   bool opened;
   int currOffset;
   int index;
};

#define MAX_OPEN_FILES (16)

struct IData
{
   IData() {reset();}
   void reset();
   bool isMounted;
   bool isInit;
   int flash_block_size;
   int flash_num_blocks;
   int fisdir_size;

   char *flash_start;
   char *flash_end;
   char* fis_addr0;
   char* fis_addr1;

   char buf0[128*1024];
   char buf1[128*1024];
   fis_valid_info* valid_fis_table;
   fis_image_desc* fisTable;
   bool is0Valid;
   bool eFis;
   FileData openFiles[MAX_OPEN_FILES];

   struct _config config;
   bool config_ok;
   bool config_init_done;
};

void IData::reset()
{
   isMounted=false;
   flash_start=0;
   flash_end=0;
   flash_block_size=0;
   flash_num_blocks=0;
   fisdir_size=0;
   fis_addr0=0;
   fis_addr1=0;
   fisTable=(fis_image_desc*)buf0;
   is0Valid=true;
   eFis=false;
   config_init_done=false;
   config_ok=false;
   for (unsigned int i=0; i<MAX_OPEN_FILES; i++)
      openFiles[i].opened=false;
}

IData g_data;

int get_valid_buf(fis_valid_info* fvi0, fis_valid_info* fvi1, int* updateWasInterrupted)
{
   CYG_ASSERT(updateWasInterrupted!=0, "updateWasInterrupted==0");
   *updateWasInterrupted=0;
   if (fvi0->magic!=EFIS_MAGIC)  //buf1 must be valid
   {
      if (fvi1->version_count>0)
         *updateWasInterrupted=1;
      return 1;
   }
   else if (fvi1->magic!=EFIS_MAGIC) //buf0 must be valid
   {
      if (fvi0->version_count>0)  //wenn noch nie ein update gelaufen ist, kann die zweite tabelle nohc 0xff sein
         *updateWasInterrupted=1;
      return 0;
   }

   //magic stimmt bei beiden, jetzt ueberpruefen, ob bei beiden valid stimmt
   if (fvi0->valid_flag!=EFIS_VALID) //buf1 must be valid
   {
      *updateWasInterrupted=1;  //wenn noch nie ein update gelaufen ist, bleiben wir schon oben bei der magic haengen
      return 1;
   }
   else if (fvi1->valid_flag!=EFIS_VALID) //buf0 must be valid
   {
      *updateWasInterrupted=1;
      return 0;
   }

   //now check the version
   if ((fvi1->version_count > fvi0->version_count) //buf1 must be valid
       && (fvi1->version_count != 0xffffffff))
      return 1;
//   else
   return 0;
}

int mount()
{
   void* err_addr=0;

   if (g_data.isMounted)
      return EBUSY;
   CYG_TRACE0(debugFisFS, "");

   if (!g_data.isInit)
   {
      g_data.isInit=true;
      CYG_TRACE0(debugFisFS, "calling flash_init");
      if (flash_init(diag_printf)!=0)
      {
         diag_printf("flash_init() failed\n");
         return EIO;
      }
   }

   flash_get_limits((void *)0, (void **)&g_data.flash_start, (void **)&g_data.flash_end);
   // Keep 'end' address as last valid location, to avoid wrap around problems
   g_data.flash_end = g_data.flash_end - 1;
   flash_get_block_info(&g_data.flash_block_size, &g_data.flash_num_blocks);
   g_data.fisdir_size=g_data.flash_block_size; // fuers erste fest auf 1 block

   CYG_TRACE4(debugFisFS, "flash_start: %p flash_end: %p flash_block_size: %d flash_num_blocks: %d", g_data.flash_start, g_data.flash_end, g_data.flash_block_size, g_data.flash_num_blocks);
   CYG_TRACE3(debugFisFS, "fisdir_size: %d CYGNUM_REDBOOT_FIS_DIRECTORY_ENTRY_COUNT: %d CYGNUM_REDBOOT_FIS_DIRECTORY_ENTRY_SIZE: %d",
               g_data.fisdir_size, CYGNUM_REDBOOT_FIS_DIRECTORY_ENTRY_COUNT, CYGNUM_REDBOOT_FIS_DIRECTORY_ENTRY_SIZE);

   if (CYGNUM_REDBOOT_FIS_DIRECTORY_BLOCK < 0)
   { //von hinten zaehlen
      g_data.fis_addr0 = g_data.flash_end + 1 +
         (CYGNUM_REDBOOT_FIS_DIRECTORY_BLOCK*g_data.flash_block_size);

//      g_data.fis_addr1 = g_data.flash_end + 1 +
//         ((CYGNUM_REDBOOT_FIS_DIRECTORY_BLOCK+1)*g_data.flash_block_size);
   }
   else
   { //von vorne zaehlen
      g_data.fis_addr0 = g_data.flash_start +
         (CYGNUM_REDBOOT_FIS_DIRECTORY_BLOCK*g_data.flash_block_size);
//      g_data.fis_addr1 = g_data.flash_start +
//         ((CYGNUM_REDBOOT_FIS_DIRECTORY_BLOCK+1)*g_data.flash_block_size);
   }
   g_data.fis_addr1 = g_data.fis_addr0+g_data.flash_block_size;

   CYG_TRACE1(debugFisFS, "reading 0 from %p", g_data.fis_addr0);

   flash_read(g_data.fis_addr0, g_data.buf0, g_data.fisdir_size, (void **)&err_addr);
   flash_read(g_data.fis_addr1, g_data.buf1, g_data.fisdir_size, (void **)&err_addr);

   fis_valid_info* fvi0=(fis_valid_info*)g_data.buf0;
   fis_valid_info* fvi1=(fis_valid_info*)g_data.buf1;

   CYG_TRACE2(debugFisFS, "addr0: %p addr1: %p", g_data.fis_addr0, g_data.fis_addr1);
   CYG_TRACE3(debugFisFS, "0: magic: 0x%x version_count: %d valid: 0x%x", fvi0->magic, fvi0->version_count, fvi0->valid_flag);
   CYG_TRACE3(debugFisFS, "1: magic: 0x%x version_count: %d valid: 0x%x", fvi1->magic, fvi1->version_count, fvi1->valid_flag);

   int updateWasInterrupted=0;
   //now check for validity
   if ((fvi0->magic==EFIS_MAGIC)
       || (fvi1->magic==EFIS_MAGIC))  //enhanced fis, select the valid one
   {
      g_data.eFis=true;
      if (get_valid_buf(fvi0, fvi1, &updateWasInterrupted)==0)
      {
         g_data.is0Valid=true;
         g_data.fisTable=(fis_image_desc*)g_data.buf0;
      }
      else
      {
         g_data.is0Valid=false;
         g_data.fisTable=(fis_image_desc*)g_data.buf1;
      }
   }
   CYG_TRACE2(debugFisFS, "efis: %d valid: %d", g_data.eFis?1:0, g_data.is0Valid?0:1);
   g_data.isMounted=true;
   if (updateWasInterrupted)
      return EXDEV;

   return 0;
}

void init()
{
   if (!g_data.isMounted)
   {
      CYG_TRACE0(debugFisFS, "not mounted");
      return;
   }

   fis_valid_info* fvi0=(fis_valid_info*)g_data.buf0;

   fvi0->magic=EFIS_MAGIC;
   fvi0->valid_flag=EFIS_VALID;
   fvi0->version_count=0;

   finishUpdateDirectory(true);

   umount();
   mount();
}

int umount()
{
   if (!g_data.isMounted)
      return EINVAL;

   g_data.reset();
   return 0;

}

fis_image_desc* find_entry(const char* name, int* index)
{
   name=name+1; //skip the "/"
   for (unsigned int i=0; i<CYGNUM_REDBOOT_FIS_DIRECTORY_ENTRY_COUNT; i++)
   {
      if (g_data.fisTable[i].name[0]==0xff)
         continue;

      if (strcmp((const char*)g_data.fisTable[i].name, name)==0)
      {
         if (index!=0)
            *index=i;
         return &g_data.fisTable[i];
      }
   }
   return 0;
}

struct fis_image_desc* getEntry(const char* name)
{
   return find_entry(name, 0);
}



int open(const char* name)
{
   if (!g_data.isMounted)
      return ENODEV;

   int fd=-1;

   for (unsigned int i=0; i<MAX_OPEN_FILES; i++)
   {
      if (g_data.openFiles[i].opened==false)
      {
         fd=i;
         break;
      }
   }
   if (fd==-1) //nix frei
      return EMFILE;

   int index=0;
   fis_image_desc* img=find_entry(name, &index);
   if (img==0)
      return ENOENT;

   g_data.openFiles[fd].opened=true;
   g_data.openFiles[fd].currOffset=0;
   g_data.openFiles[fd].index=index;
   return fd;
}

#define MIN(a,b) (((a)<(b))?(a):(b))

int read(int fd, char* buf, int size)
{
   if (!g_data.isMounted)
      return ENODEV;

   if (( fd >= MAX_OPEN_FILES) || (!g_data.openFiles[fd].opened))
      return EBADF;

   void* err_addr=0;

   int readBytes=MIN(size, g_data.fisTable[g_data.openFiles[fd].index].data_length-g_data.openFiles[fd].currOffset);
   if (readBytes<1)
      return 0;

   flash_read((void *)(g_data.fisTable[g_data.openFiles[fd].index].flash_base+g_data.openFiles[fd].currOffset), buf, readBytes, (void **)&err_addr);

   g_data.openFiles[fd].currOffset+=readBytes;
   return readBytes;
}

int close(int fd)
{
   if (!g_data.isMounted)
      return ENODEV;

   if (( fd >= MAX_OPEN_FILES) || (!g_data.openFiles[fd].opened))
      return EBADF;

   g_data.openFiles[fd].opened=false;
   g_data.openFiles[fd].currOffset=0;
   g_data.openFiles[fd].index=0;
   return 0;
}

int opendir(const char* name, Dir* dir)
{
   if (!g_data.isMounted)
      return ENODEV;

   CYG_TRACE2(debugFisFS, "efis: %d valid: %d", g_data.eFis?1:0, g_data.is0Valid?0:1);
   fis_valid_info* fvi0=(fis_valid_info*)g_data.buf0;
   fis_valid_info* fvi1=(fis_valid_info*)g_data.buf1;

   CYG_TRACE3(debugFisFS, "0: magic: 0x%x valid: 0x%x version_count: %d", fvi0->magic, fvi0->valid_flag, fvi0->version_count);
   CYG_TRACE3(debugFisFS, "1: magic: 0x%x valid: 0x%x version_count: %d", fvi1->magic, fvi1->valid_flag, fvi1->version_count);

   if (strcmp(name, "/")!=0)
      return ENOENT;
   dir->current_pos=0;
   return 0;
}

int readdir(Dir* dir, DirEntry* dirEntry)
{
   if (!g_data.isMounted)
      return ENODEV;

   while (g_data.fisTable[dir->current_pos].name[0]==0xff)
   {
      dir->current_pos++;
      if (dir->current_pos>=CYGNUM_REDBOOT_FIS_DIRECTORY_ENTRY_COUNT)
         return EEOF; /// \todo
   }

   strcpy(dirEntry->name, (const char*)g_data.fisTable[dir->current_pos].name);
   dirEntry->size=g_data.fisTable[dir->current_pos].data_length;
   dir->current_pos++;
   return 0;
}

int closeDir(Dir* dir)
{
   if (!g_data.isMounted)
      return ENODEV;

   dir->current_pos=0;
   return 0;
}

void readDirFromFlash()
{
   void * err_addr=0;
   flash_read(g_data.fis_addr0, g_data.buf0, g_data.fisdir_size, (void **)&err_addr);
   flash_read(g_data.fis_addr1, g_data.buf1, g_data.fisdir_size, (void **)&err_addr);
}

int eraseImage(const char* name)
{
   if (!g_data.isMounted)
   {
      CYG_TRACE0(debugFisFS, "not mounted");
      return ENODEV;
   }

   fis_image_desc* img=getEntry(name);
   if (img==0)
   {
      CYG_TRACE1(debugFisFS, "no such image %s", name);
      return ENOENT;
   }

   if ((strcmp(name+1, "FIS directory")==0)
       || (strcmp(name+1, "RedBoot config")==0))
   {
      CYG_TRACE1(debugFisFS, "deleting %s not permitted", name);
      return EPERM;
   }

   //modify the directory so that it reflects the changes in RAM
   fis_image_desc imgBackup;
   imgBackup=*img;

   memset(img, 0xff, sizeof(fis_image_desc));

   CYG_TRACE3(debugFisFS, "erasing -%s- addr: %x size: %d", name, imgBackup.flash_base, imgBackup.size);

   //write them for now into the flash
   startUpdateDirectory();
   //erase it
   void *err_addr=0;
   flash_unlock((void *)imgBackup.flash_base, imgBackup.size, (void **)&err_addr);

   flash_erase((void *)imgBackup.flash_base, imgBackup.size, (void **)&err_addr);

   flash_lock((void *)imgBackup.flash_base, imgBackup.size, (void **)&err_addr);
   //and make the new written directory valid
   finishUpdateDirectory(true);
   return 0;
}

//create a new image, fails if it already exists
int createImage(const char* name, char* buf,
                CYG_ADDRESS flash_base, CYG_ADDRESS mem_base, unsigned long size,
                CYG_ADDRESS entry_point, unsigned long data_length/*, const char* replacingName*/, bool backup)
{
   if (!g_data.isMounted)
   {
      CYG_TRACE0(debugFisFS, "not mounted");
      return ENODEV;
   }

   //check parameters
   if (((size % g_data.flash_block_size) !=0) || (size==0))
   {
      CYG_TRACE1(debugFisFS, "invalid size: %d", size);
      return EINVAL;
   }

   if ((flash_verify_addr((void*)flash_base)!=0)
       || (flash_verify_addr((void*)(flash_base+size-1))!=0))
   {
      CYG_TRACE2(debugFisFS, "invalid addresses %p %p", flash_base, flash_base+size-1);
      return EINVAL;
   }

   char backupName[17];
   sprintf(backupName, "%s~", name);

   fis_image_desc* oldReplacedImg=0;
   fis_image_desc* existingImg=getEntry(name); //which might be replaced
   if (!backup)  //create new, don't replace
   {
      if (existingImg!=0)
      {
         CYG_TRACE1(debugFisFS, "image already exists: %s", name);
         return EEXIST;
      }
   }
   else
   {
      fis_image_desc* oldReplacedImg=getEntry(backupName);
      if (oldReplacedImg!=0)   //this will go away
      {
         oldReplacedImg->name[0]=0xff;
         oldReplacedImg->name[1]=0xff;
      }
   }

   fis_image_desc* img=0;
   //find a free entry
   for (unsigned int i=0; i<CYGNUM_REDBOOT_FIS_DIRECTORY_ENTRY_COUNT; i++)
   {
      if ((g_data.fisTable[i].name[0]==0xff)
          && (g_data.fisTable[i].name[1]==0xff))
      {
         img=&g_data.fisTable[i];
         break;
      }
   }


   if (img==0)
   {
      CYG_TRACE0(debugFisFS, "no free slot left");
      if (oldReplacedImg!=0)   //restore the entry
      {
         oldReplacedImg->name[0]=backupName[1];
         oldReplacedImg->name[1]=backupName[2];
      }
      return EMFILE;
   }

   char usedBlocks[BLOCK_NUMBER];
   memset(usedBlocks, 0, BLOCK_NUMBER);
   for (unsigned int i=0; i<CYGNUM_REDBOOT_FIS_DIRECTORY_ENTRY_COUNT; i++)
   {
      if (g_data.fisTable[i].name[0]==0xff)
         continue;

      int startBlock=(g_data.fisTable[i].flash_base-(CYG_ADDRESS)g_data.flash_start) / g_data.flash_block_size;
      int endBlock=((g_data.fisTable[i].flash_base-(CYG_ADDRESS)g_data.flash_start) + g_data.fisTable[i].size - 1) / g_data.flash_block_size;

      for (int j=startBlock; j<=endBlock; j++)
         usedBlocks[j]=1;
   }

   if (debugFisFS)
      diag_dump_buf(usedBlocks, BLOCK_NUMBER);

   int startBlock=(flash_base-(CYG_ADDRESS)g_data.flash_start) / g_data.flash_block_size;
   int endBlock=((flash_base-(CYG_ADDRESS)g_data.flash_start) + size - 1) / g_data.flash_block_size;

   CYG_TRACE2(debugFisFS, "using now from %d to %d", startBlock, endBlock);

   for (int i=startBlock; i<=endBlock; i++)
   {
      if (usedBlocks[i])
      {
         CYG_TRACE1(debugFisFS, "block %d is used", i);
         if (oldReplacedImg!=0)   //restore the entry
         {
            oldReplacedImg->name[0]=backupName[1];
            oldReplacedImg->name[1]=backupName[2];
         }
         return ENOMEM;
      }
   }

   if (backup)
   {
      if (oldReplacedImg!=0)  //throw the old backup away
         memset(oldReplacedImg, 0xff, sizeof(fis_image_desc));

      if (existingImg!=0)
      {
         strcpy((char*)existingImg->name, backupName+1);
      }
   }

   //modify the directory so that it reflects the changes in RAM
   memset(img, 0xff, sizeof(fis_image_desc));
   strcpy((char*)img->name, name+1);
   img->flash_base=flash_base;
   img->mem_base=mem_base;
   img->size=size;
   img->entry_point=entry_point;
   img->data_length=data_length;
   img->file_cksum=cyg_crc32((unsigned char*)buf, data_length);

   //write them for now into the flash
   if (startUpdateDirectory()!=0)
   {
      // copy im ram wieder in ordnung bringen....
      CYG_TRACE0(debugFisFS, "files open for reading");
      readDirFromFlash();
      return -1;
   }

   //unlock, erase, program and lock  it
   void *err_addr=0;
   CYG_TRACE1(debugFisFS, "unlocking %p", img->flash_base);
   flash_unlock((void *)img->flash_base, img->size, (void **)&err_addr);

   CYG_TRACE1(debugFisFS, "erasing %p", img->flash_base);
   flash_erase((void *)img->flash_base, img->size, (void **)&err_addr);

   CYG_TRACE2(debugFisFS, "programming %p, size: %d", img->flash_base, img->data_length);
   flash_program((void *)img->flash_base, (void*)buf, img->data_length, (void **)&err_addr);

   CYG_TRACE1(debugFisFS, "locking %p", img->flash_base);
   flash_lock((void *)img->flash_base, img->size, (void **)&err_addr);

   //read back and check the crc
   flash_read((void *)img->flash_base, buf, img->data_length, (void **)&err_addr);
//   memset(buf, 0, 1000);  //check that crc checking works
   unsigned long cksum=cyg_crc32((unsigned char*)buf, data_length);
   if (img->file_cksum!=cksum)
   {
      CYG_TRACE2(debugFisFS, "crc error: read: 0x%x expected: 0x%x", cksum, img->file_cksum);
      finishUpdateDirectory(false);
      return -1;
   }


   //and make the new written directory valid
   finishUpdateDirectory(true);

   return 0;
}

/*int createImage(const char* name, const char* buf,
                CYG_ADDRESS flash_base, CYG_ADDRESS mem_base, unsigned long size,
                CYG_ADDRESS entry_point, unsigned long data_length)
{
   return _createImage(name, buf, flash_base, mem_base, size, entry_point, data_length, 0);
} */


/*special purpose: create a new image, if it already exists, rename the existing one
 to oldName, if oldName already exists, throw it away */
/*int exchangeImages(const char* newName, const char* oldName, const char* buf,
                   CYG_ADDRESS flash_base, CYG_ADDRESS mem_base, unsigned long size,
                   CYG_ADDRESS entry_point, unsigned long data_length, bool backup)
{
   return _createImage(newName, buf, flash_base, mem_base, size, entry_point, data_length, oldName);
}*/

int startUpdateDirectory()
{
   for (unsigned i=0; i<MAX_OPEN_FILES; i++)
   {
      if (g_data.openFiles[i].opened)
         return EINVAL;
   }

   if (!g_data.eFis) //nothing to do with old fis...
      return 0;


   const char* src=0;
   char* dest=0;
   void* fis_addr=0;
   if (g_data.is0Valid)  //jetzt wird 1 valid
   {
      src=g_data.buf0;
      dest=g_data.buf1;
      fis_addr = g_data.fis_addr1;
   }
   else
   {
      src=g_data.buf1;
      dest=g_data.buf0;
      fis_addr = g_data.fis_addr0;
   }
   g_data.is0Valid=!g_data.is0Valid;

   memcpy(dest, src, g_data.fisdir_size);
   fis_valid_info* new_fvi=(fis_valid_info*)dest;
   const fis_valid_info* old_fvi=(const fis_valid_info*)src;

   new_fvi->magic=EFIS_MAGIC;
   new_fvi->valid_flag=EFIS_IN_PROGRESS;
   new_fvi->version_count=old_fvi->version_count+1;

   //ready to go....

   void* err_addr=0;
   flash_unlock((void *)fis_addr, g_data.flash_block_size, (void **)&err_addr);
   flash_erase(fis_addr, g_data.flash_block_size, (void **)&err_addr);
   //now magic is 0xffffffff
   flash_program(fis_addr, dest, g_data.flash_block_size, (void **)&err_addr);
   //now magic is 0xff1234ff, valid is IN_PROGRESS, version_count is the old one +1
   return 0;
}

int finishUpdateDirectory(bool success)
{
   if (!g_data.eFis)
   {
      void* err_addr=0;
      CYG_TRACE2(debugFisFS, "unlocking %p, size: %d", g_data.fis_addr0, g_data.flash_block_size);
      flash_unlock((void *)g_data.fis_addr0, g_data.flash_block_size, (void **)&err_addr);
      flash_erase(g_data.fis_addr0, g_data.flash_block_size, (void **)&err_addr);
      //now magic is 0xffffffff
      flash_program(g_data.fis_addr0, g_data.buf0, g_data.flash_block_size, (void **)&err_addr);
      flash_lock((void *)g_data.fis_addr0, g_data.flash_block_size, (void **)&err_addr);
      return 0;
   }

   void* err_addr=0;

   char* dest=0;
   void* fis_addr=0;
   if (g_data.is0Valid)  //jetzt ist 0 aktiv
   {
      dest=g_data.buf0;
      fis_addr = g_data.fis_addr0;
   }
   else
   {
      dest=g_data.buf1;
      fis_addr = g_data.fis_addr1;
   }

   //if it failed, reset is0Valid to the state before startUpdateDirectory()
   //g_data.fisTable hans't been changed yet, so it doesn't have to be reset now
   //then reread the contents from flash
   //setting the valid flag of the failed table to "INAVLID" might also be not too bad
   //but IN_PROGRESS is also good enough I think
   if (!success)
   {
      g_data.is0Valid=!g_data.is0Valid;
      readDirFromFlash();
   }
   else //success
   {
      g_data.fisTable=(fis_image_desc*)dest;
      fis_addr = (void *)((CYG_ADDRESS)fis_addr+sizeof(unsigned int));

      fis_valid_info* new_fvi=(fis_valid_info*)dest;
      new_fvi->valid_flag=EFIS_VALID;

      CYG_TRACE1(debugFisFS, "programming %p", fis_addr);
      flash_program(fis_addr, (char*)&new_fvi->valid_flag, sizeof(new_fvi->valid_flag), (void **)&err_addr);

      //ab jetzt sollte das neue directory gueltig sein
      fis_addr = (void *)((CYG_ADDRESS)fis_addr-sizeof(unsigned int));
   }

   flash_lock((void *)fis_addr, g_data.flash_block_size, (void **)&err_addr);
   return 0;
}


static int
config_length(int type)
{
    switch (type) {
    case CONFIG_BOOL:
       return 4;  //im redboot (in C) ist sizeof(bool)==4, hier (in C++) ist sizeof(bool)==1
//        return sizeof(bool);
    case CONFIG_INT:
        return sizeof(unsigned long);
#ifdef CYGPKG_REDBOOT_NETWORKING*/
    case CONFIG_IP:
//       CYG_TRACE0(debugFisFS, "CONFIG_IP");
//       return 4;
        return sizeof(in_addr_t);
    case CONFIG_ESA:
        // Would like this to be sizeof(enet_addr_t), but that causes much
        // pain since it fouls the alignment of data which follows.
//       CYG_TRACE0(debugFisFS, "CONFIG_ESA");
        return 8;
#if defined(CYGHWR_NET_DRIVERS) && (CYGHWR_NET_DRIVERS > 1)
    case CONFIG_NETPORT:
//       CYG_TRACE0(debugFisFS, "CONFIG_NETPORT");

        return MAX_STRING_LENGTH;
#endif
#endif
    case CONFIG_STRING:
        return MAX_STRING_LENGTH;
    case CONFIG_SCRIPT:
        return MAX_SCRIPT_LENGTH;
    default:
        return 0;
    }
}

//
// Find the configuration entry for a particular key
//
unsigned char * flash_lookup_config(const char *key)
{
   unsigned char *dp;
   int len;

   if (!g_data.isMounted)
      return NULL;

   if (!g_data.config_init_done)
   {
      g_data.config_init_done=true;
      g_data.config_ok=false;

      fis_image_desc* img=getEntry("/RedBoot config");
      if (img!=0)
      {
         void *err_addr=0;
         flash_read((void *)img->flash_base, &g_data.config, sizeof(_config), (void **)&err_addr);

         if ((g_data.config.key1==CONFIG_KEY1) || (g_data.config.key2==CONFIG_KEY2))
            g_data.config_ok=true;
         //crc ueberpruefen
      }
      else
      {
         CYG_TRACE0(debugFisFS, "image not found");
      }
   }

   if (!g_data.config_ok) return (unsigned char *)NULL;

   CYG_TRACE1(debugFisFS, "config at: %p", &g_data.config);
   dp = &g_data.config.config_data[0];
   while (dp < &g_data.config.config_data[sizeof(g_data.config.config_data)]) {
      len = 4 + CONFIG_OBJECT_KEYLEN(dp) + CONFIG_OBJECT_ENABLE_KEYLEN(dp) +
         config_length(CONFIG_OBJECT_TYPE(dp));
      CYG_TRACE5(debugFisFS, "key: -%s- dp: %p keylen: %d, key_enable_len: %d cfg length: %d", (const char*)CONFIG_OBJECT_KEY(dp), dp, CONFIG_OBJECT_KEYLEN(dp), CONFIG_OBJECT_ENABLE_KEYLEN(dp), config_length(CONFIG_OBJECT_TYPE(dp)));
      if (strcmp(key, (const char*)CONFIG_OBJECT_KEY(dp)) == 0) {
         return (unsigned char*)CONFIG_OBJECT_VALUE(dp);
      }
      dp += len;
   }
//    diag_printf("Can't find config data for '%s'\n", key);
   return NULL;
}


const char* getFConfig(const char* key, const char* defaultValue)
{
   char tmp[256];
   sprintf(tmp, "alias/%s", key);
   const char* value=(const char*)flash_lookup_config(tmp);
   if (value==0)
      return defaultValue;
   return value;
}


int checkChecksum(const char* name, unsigned int workBufSize, char* workBuf)
{
   struct fis_image_desc* img=getEntry(name);
   if (img==0)
      return ENOENT;

   if (workBufSize<img->data_length)
      return ENOMEM;

   void* err_addr=0;
   flash_read((void *)img->flash_base, workBuf, img->data_length, (void **)&err_addr);

   unsigned long file_cksum = cyg_crc32((unsigned char*)workBuf, img->data_length);
   CYG_TRACE4(debugFisFS, "%s length: %d cksum: %x expected cksum: %x", name, img->data_length, file_cksum, img->file_cksum);
   if (file_cksum!=img->file_cksum)
      return -1;

   return 0;
}


};



