#ifndef FISFS_H
#define FISFS_H

#include "should_come_from_configtool.h"
#include "fis.h"

namespace FisFS
{

unsigned char * flash_lookup_config(const char *key);

const char* getFConfig(const char* key, const char* defaultValue=0);

void init();

/**mount the fis fs. Returns 0 on success. Returns
 EXDEV if the mounting succeeded but an interrupted write operation
 was detected. In this case you can use the file system, but you should
 check that all files are still valid, e.g. using checkChecksum().
 */
int mount();
int umount();

int open(const char* name);

int read(int fd, char* buf, int size);

int close(int fd);

struct DirEntry
{
   char name[16];
   int size;
};

struct Dir
{
   int current_pos;
};

int opendir(const char* name, Dir* dir);

int readdir(Dir* dir, DirEntry* dirEntry);

int closeDir(Dir* dir);

///erase an existing image, fails if it doesn't exist
int eraseImage(const char* name);

/** create a new image, fails if it already exists and backup is false, if backup is true,
 then the existing image with this name will get an ~ appended. If a previous backup file
 already exists it will be removed.*/
int createImage(const char* name, char* buf,
                CYG_ADDRESS flash_base, CYG_ADDRESS mem_base, unsigned long size,
                CYG_ADDRESS entry_point, unsigned long data_length, bool backup);


struct fis_image_desc* getEntry(const char* name);

int checkChecksum(const char* name, unsigned int workBufSize, char* workBuf);

};

extern bool debugFisFS;


#endif


